﻿using System;
using System.Collections.Generic;
using System.Text;

namespace S3Delete
{
    class Program
    {
        private static String S3FolderDelimiter = "/";
        private static int ExitCode;
        private static String EmailLogData;
        private static int OutputLevel;
        private static String LogFilePath;
        private static System.IO.StreamWriter LogStreamWriter;

        private static int EmailLogCondition;
        private static String EmailSmtpHost;
        private static int EmailSmtpPort;
        private static Boolean EmailSmtpEnableSsl;
        private static String EmailSmtpUserName;
        private static String EmailSmtpPassword;
        private static String EmailFromAddress;
        private static String EmailRecipients;

        private static Boolean DeleteSubFolders;
        private static Boolean DeleteSpecifiedFolder;
        private static String IncludeOnlyS3FilesRegularExpression;
        private static String ExcludeS3FilesRegularExpression;
        private static Double DeleteFilesOlderThanDays;
        private static DateTime ProgressUpdateShown = DateTime.MinValue;

        
        static void Main(string[] args)
        {
            AppDomain currentDomain = AppDomain.CurrentDomain;
            currentDomain.UnhandledException += new UnhandledExceptionEventHandler(MyUnhandledException);

            ExitCode = 0;
            EmailLogData = "";

            Boolean LogOnlyMode;
            if (Boolean.TryParse(GetValueFromConfigOrArgument(args, "LogOnlyMode"), out LogOnlyMode) == false)
            {
                LogOnlyMode = true;
            }

            if (int.TryParse(GetValueFromConfigOrArgument(args, "OutputLevel"), out OutputLevel) == false)
            {
                OutputLevel = 1;
            }

            LogFilePath = GetValueFromConfigOrArgument(args, "LogFilePath");

            Boolean OverwriteExistingLogFile;
            if (Boolean.TryParse(GetValueFromConfigOrArgument(args, "OverwriteExistingLogFile"), out OverwriteExistingLogFile) == false)
            {
                OverwriteExistingLogFile = true;
            }

            String AWSAccessKeyId = GetValueFromConfigOrArgument(args, "AWSAccessKeyId");
            String AWSSecretAccessKey = GetValueFromConfigOrArgument(args, "AWSSecretAccessKey");

            Boolean UseSSL;
            if (Boolean.TryParse(GetValueFromConfigOrArgument(args, "UseSSL"), out UseSSL) == false)
            {
                UseSSL = true;
            }

            String RequestEndpoint = GetValueFromConfigOrArgument(args, "RequestEndpoint");

            String BucketName = GetValueFromConfigOrArgument(args, "BucketName");
            String S3KeyName = GetValueFromConfigOrArgument(args, "S3KeyName");
            String UploadHeaders = GetValueFromConfigOrArgument(args, "UploadHeaders");
            String OverrideContentTypes = GetValueFromConfigOrArgument(args, "OverrideContentTypes");

            Boolean ShowProgress;
            if (Boolean.TryParse(GetValueFromConfigOrArgument(args, "ShowProgress"), out ShowProgress) == false)
            {
                ShowProgress = true;
            }

            int S3ErrorRetries;
            if (int.TryParse(GetValueFromConfigOrArgument(args, "S3ErrorRetries"), out S3ErrorRetries) == false)
            {
                S3ErrorRetries = 10;
            }

            if (Boolean.TryParse(GetValueFromConfigOrArgument(args, "DeleteSubFolders"), out DeleteSubFolders) == false)
            {
                DeleteSubFolders = true;
            }
            if (Boolean.TryParse(GetValueFromConfigOrArgument(args, "DeleteSpecifiedFolder"), out DeleteSpecifiedFolder) == false)
            {
                DeleteSpecifiedFolder = true;
            }

            IncludeOnlyS3FilesRegularExpression = GetValueFromConfigOrArgument(args, "IncludeOnlyS3FilesRegularExpression");
            ExcludeS3FilesRegularExpression = GetValueFromConfigOrArgument(args, "ExcludeS3FilesRegularExpression");

            if (Double.TryParse(GetValueFromConfigOrArgument(args, "DeleteFilesOlderThanDays"), out DeleteFilesOlderThanDays) == false)
            {
                DeleteFilesOlderThanDays = 0;
            }

            if (int.TryParse(GetValueFromConfigOrArgument(args, "EmailLogCondition"), out EmailLogCondition) == false)
            {
                EmailLogCondition = 0;
            }
            EmailSmtpHost = GetValueFromConfigOrArgument(args, "EmailSmtpHost");
            if (int.TryParse(GetValueFromConfigOrArgument(args, "EmailSmtpPort"), out EmailSmtpPort) == false)
            {
                EmailSmtpPort = 0;
            }
            if (Boolean.TryParse(GetValueFromConfigOrArgument(args, "EmailSmtpEnableSsl"), out EmailSmtpEnableSsl) == false)
            {
                EmailSmtpEnableSsl = false;
            }
            EmailSmtpUserName = GetValueFromConfigOrArgument(args, "EmailSmtpUserName");
            EmailSmtpPassword = GetValueFromConfigOrArgument(args, "EmailSmtpPassword");
            EmailFromAddress = GetValueFromConfigOrArgument(args, "EmailFromAddress");
            EmailRecipients = GetValueFromConfigOrArgument(args, "EmailRecipients");


            //Initialize the log file.
            InitializeLogFile(OverwriteExistingLogFile);


            WriteToLog("");
            WriteToLog("Program started.   " + DateTime.Now);

            WriteToLog("", 2);
            WriteToLog("Input parameters:", 2);
            WriteToLog("   AWSAccessKeyId=" + AWSAccessKeyId, 2);
            WriteToLog("   AWSSecretAccessKey=" + AWSSecretAccessKey, 2);
            WriteToLog("   UseSSL=" + UseSSL, 2);
            WriteToLog("   RequestEndpoint=" + RequestEndpoint, 2);
            WriteToLog("   BucketName=" + BucketName, 2);
            WriteToLog("   S3KeyName=" + S3KeyName, 2);

            WriteToLog("   LogOnlyMode=" + LogOnlyMode, 2);
            WriteToLog("   OutputLevel=" + OutputLevel, 2);
            WriteToLog("   LogFilePath=" + LogFilePath, 2);
            WriteToLog("   OverwriteExistingLogFile=" + OverwriteExistingLogFile, 2);

            WriteToLog("   ShowProgress=" + ShowProgress, 2);
            WriteToLog("   S3ErrorRetries=" + S3ErrorRetries, 2);

            WriteToLog("   DeleteSubFolders=" + DeleteSubFolders, 2);
            WriteToLog("   DeleteSpecifiedFolder=" + DeleteSpecifiedFolder, 2);
            WriteToLog("   IncludeOnlyS3FilesRegularExpression=" + IncludeOnlyS3FilesRegularExpression, 2);
            WriteToLog("   ExcludeS3FilesRegularExpression=" + ExcludeS3FilesRegularExpression, 2);
            WriteToLog("   DeleteFilesOlderThanDays=" + DeleteFilesOlderThanDays, 2);

            WriteToLog("   EmailLogCondition=" + EmailLogCondition, 2);
            WriteToLog("   EmailSmtpHost=" + EmailSmtpHost, 2);
            WriteToLog("   EmailSmtpPort=" + EmailSmtpPort, 2);
            WriteToLog("   EmailSmtpEnableSsl=" + EmailSmtpEnableSsl, 2);
            WriteToLog("   EmailSmtpUserName=" + EmailSmtpUserName, 2);
            WriteToLog("   EmailSmtpPassword=" + EmailSmtpPassword, 2);
            WriteToLog("   EmailFromAddress=" + EmailFromAddress, 2);
            WriteToLog("   EmailRecipients=" + EmailRecipients, 2);
            WriteToLog("", 2);


            //Check that the input values are valid.
            if (AWSAccessKeyId == "")
            {
                WriteToLog("Error, the AWSAccessKeyId is empty.  Set it in the .config file or pass it in the command line.");
                ExitCode = -1;
            }
            else if (AWSSecretAccessKey == "")
            {
                WriteToLog("Error, the AWSSecretAccessKey is empty.  Set it in the .config file or pass it in the command line.");
                ExitCode = -1;
            }
            else if (BucketName == "")
            {
                WriteToLog("Error, the BucketName is empty.  Set it in the .config file or pass it in the command line.");
                ExitCode = -1;
            }

            SprightlySoftAWS.REST MyREST = new SprightlySoftAWS.REST();
            String RequestURL;
            String RequestMethod;
            Boolean RetBool;
            int ErrorNumber = 0;
            String ErrorDescription = "";
            String LogData = "";
            Dictionary<String, String> RequestHeaders = new Dictionary<String, String>();
            int ResponseStatusCode = 0;
            String ResponseStatusDescription = "";
            Dictionary<String, String> ResponseHeaders = new Dictionary<String, String>();
            String ResponseString = "";
            String ResponseErrorMessage = "";

            if (ExitCode == 0)
            {
                //Check if the account info is valid.
                //Get the bucket location.

                //GET Bucket location: http://docs.amazonwebservices.com/AmazonS3/latest/API/RESTBucketLocationGET.html
                //Regions and Endpoints:  http://docs.amazonwebservices.com/general/latest/gr/index.html?rande.html

                RequestURL = MyREST.BuildS3RequestURL(UseSSL, "s3.amazonaws.com", BucketName, "", "?location");
                RequestMethod = "GET";

                RetBool = MakeS3Request(AWSAccessKeyId, AWSSecretAccessKey, RequestURL, RequestMethod, null, "", S3ErrorRetries, ref ErrorNumber, ref ErrorDescription, ref LogData, ref RequestHeaders, ref ResponseStatusCode, ref ResponseStatusDescription, ref ResponseHeaders, ref ResponseString);

                if (RetBool == true)
                {
                    if (RequestEndpoint == "")
                    {
                        System.Xml.XmlDocument XmlDoc;
                        System.Xml.XmlNamespaceManager MyXmlNamespaceManager;
                        System.Xml.XmlNode XmlNode;

                        XmlDoc = new System.Xml.XmlDocument();
                        XmlDoc.LoadXml(ResponseString);

                        MyXmlNamespaceManager = new System.Xml.XmlNamespaceManager(XmlDoc.NameTable);
                        MyXmlNamespaceManager.AddNamespace("amz", "http://s3.amazonaws.com/doc/2006-03-01/");

                        XmlNode = XmlDoc.SelectSingleNode("/amz:LocationConstraint", MyXmlNamespaceManager);

                        if (XmlNode.InnerText == "")
                        {
                            RequestEndpoint = "s3.amazonaws.com";
                        }
                        else if (XmlNode.InnerText == "us-west-1")
                        {
                            RequestEndpoint = "s3-us-west-1.amazonaws.com";
                        }
                        else if (XmlNode.InnerText == "EU")
                        {
                            RequestEndpoint = "s3-eu-west-1.amazonaws.com";
                        }
                        else if (XmlNode.InnerText == "ap-southeast-1")
                        {
                            RequestEndpoint = "s3-ap-southeast-1.amazonaws.com";
                        }
                        else if (XmlNode.InnerText == "ap-northeast-1")
                        {
                            RequestEndpoint = "s3-ap-northeast-1.amazonaws.com";
                        }
                        else
                        {
                            WriteToLog("Error, the RequestEndpoint could not be determined.  LocationConstraint=" + XmlNode.InnerText + "  Set RequestEndpoint in the config file or pass it in the command line.  See http://docs.amazonwebservices.com/general/latest/gr/rande.html#s3_region for a list of endpoints.");
                            ExitCode = -1;
                        }

                        if (RequestEndpoint != "")
                        {
                            WriteToLog("RequestEndpoint=" + RequestEndpoint, 2);
                        }
                    }
                }
                else
                {
                    WriteToLog(LogData, 2);
                    ResponseErrorMessage = FormatErrorData(RequestURL, RequestMethod, RequestHeaders, ResponseStatusCode, ResponseStatusDescription, ResponseHeaders, ResponseString, ErrorNumber, ErrorDescription);
                    WriteToLog(ResponseErrorMessage);
                    ExitCode = -1;
                }
            }


            if (ExitCode == 0)
            {
                if (S3KeyName != "")
                {
                    //Check if the S3KeyName exists.

                    RequestURL = MyREST.BuildS3RequestURL(UseSSL, RequestEndpoint, BucketName, S3KeyName, "");
                    RequestMethod = "HEAD";

                    RetBool = MakeS3Request(AWSAccessKeyId, AWSSecretAccessKey, RequestURL, RequestMethod, null, "", S3ErrorRetries, ref ErrorNumber, ref ErrorDescription, ref LogData, ref RequestHeaders, ref ResponseStatusCode, ref ResponseStatusDescription, ref ResponseHeaders, ref ResponseString);

                    if (RetBool == true)
                    {
                        //the S3KeyName exists
                    }
                    else if (ResponseStatusCode == 404)
                    {
                        WriteToLog("Error, the S3KeyName object does not exist on S3.  S3KeyName=" + S3KeyName);
                        ExitCode = -1;
                    }
                    else
                    {
                        WriteToLog(LogData, 2);
                        ResponseErrorMessage = FormatErrorData(RequestURL, RequestMethod, RequestHeaders, ResponseStatusCode, ResponseStatusDescription, ResponseHeaders, ResponseString, ErrorNumber, ErrorDescription);
                        WriteToLog(ResponseErrorMessage);
                        ExitCode = -1;
                    }
                }
            }


            if (ExitCode == 0)
            {
                System.Collections.Generic.List<String> S3DeleteList = new System.Collections.Generic.List<String>();

                if (S3KeyName.EndsWith(S3FolderDelimiter) == true | S3KeyName == "")
                {
                    WriteToLog("Getting items to delete on S3.");
                    PopulateS3DeleteList(AWSAccessKeyId, AWSSecretAccessKey, UseSSL, RequestEndpoint, BucketName, S3KeyName, S3ErrorRetries, ref S3DeleteList);
                }
                else
                {
                    S3DeleteList.Add(S3KeyName);
                }


                if (ExitCode == 0)
                {
                    if (LogOnlyMode == true)
                    {
                        WriteToLog("The following items would be deleted on S3:");

                        foreach (String DeleteItem in S3DeleteList)
                        {
                            WriteToLog("   " + DeleteItem);
                        }
                        WriteToLog("Number of items that would be deleted: " + S3DeleteList.Count);
                    }
                    else
                    {
                        //build a list of XMLs to submit for delete
                        System.Collections.Generic.List<String> XMLList = new System.Collections.Generic.List<String>();
                        String XMLRequest;
                        int ItemCounter = 0;

                        //start a new xml
                        XMLRequest = "";
                        XMLRequest += "<?xml version=\"1.0\" encoding=\"UTF-8\"?>";
                        XMLRequest += "<Delete>";
                        //XMLRequest += "<Quiet>true</Quiet>";

                        foreach (String DeleteItem in S3DeleteList)
                        {
                            ItemCounter += 1;

                            XMLRequest += "<Object><Key>" + new System.Xml.Linq.XText(DeleteItem).ToString() + "</Key></Object>";

                            if (ItemCounter == 1000)
                            {
                                //close the XML
                                XMLRequest += "</Delete>";

                                //add it to a list
                                XMLList.Add(XMLRequest);

                                //start a new xml
                                XMLRequest = "";
                                XMLRequest += "<?xml version=\"1.0\" encoding=\"UTF-8\"?>";
                                XMLRequest += "<Delete>";
                                //XMLRequest += "<Quiet>true</Quiet>";

                                ItemCounter = 0;
                            }
                        }

                        //close the XML
                        XMLRequest += "</Delete>";

                        //don't add the XML if no items in the XML
                        if (ItemCounter > 0)
                        {
                            //add it to a list
                            XMLList.Add(XMLRequest);
                        }


                        int PageCounter = 0;
                        int DeleteCounter = 0;

                        if (LogFilePath != "")
                        {
                            //Open a StreamWriter to the log file.  The WriteToLog function does not work for concurrent calls.
                            LogStreamWriter = new System.IO.StreamWriter(LogFilePath, true, System.Text.Encoding.Unicode);
                        }

                        System.Threading.CancellationTokenSource MyCancellationTokenSource = new System.Threading.CancellationTokenSource();

                        System.Threading.Tasks.ParallelOptions MyParallelOptions;
                        MyParallelOptions = new System.Threading.Tasks.ParallelOptions();
                        MyParallelOptions.CancellationToken = MyCancellationTokenSource.Token;

                        //Delete all items in the list in parallel.
                        System.Threading.Tasks.Parallel.ForEach(XMLList, MyParallelOptions, (XMLItem, MyParallelLoopState) =>
                        {
                            PageCounter += 1;
                            WriteToLogStream("Deleting page: " + PageCounter.ToString("N0") + "/" + XMLList.Count.ToString("N0"));

                            String ParallelRequestURL;
                            String ParallelRequestMethod;
                            Boolean ParallelRetBool;
                            int ParallelErrorNumber = 0;
                            String ParallelErrorDescription = "";
                            String ParallelLogData = "";
                            Dictionary<String, String> ParallelRequestHeaders = new Dictionary<String, String>();
                            int ParallelResponseStatusCode = 0;
                            String ParallelResponseStatusDescription = "";
                            Dictionary<String, String> ParallelResponseHeaders = new Dictionary<String, String>();
                            String ParallelResponseString = "";
                            String ParallelResponseErrorMessage = "";
                            Dictionary<String, String> ParallelExtraRequestHeaders = new Dictionary<String, String>();
                            String ParallelContentMD5;
                            SprightlySoftAWS.S3.CalculateHash ParallelCalculateHash = new SprightlySoftAWS.S3.CalculateHash();

                            ParallelContentMD5 = ParallelCalculateHash.CalculateMD5FromString(XMLItem);
                            ParallelExtraRequestHeaders.Add("Content-MD5", ParallelContentMD5);

                            //Send the correct parameters to the MakeS3Request function to delete a file on S3.
                            ParallelRequestURL = MyREST.BuildS3RequestURL(UseSSL, RequestEndpoint, BucketName, "", "?delete");
                            ParallelRequestMethod = "POST";

                            ParallelRetBool = MakeS3Request(AWSAccessKeyId, AWSSecretAccessKey, ParallelRequestURL, ParallelRequestMethod, ParallelExtraRequestHeaders, XMLItem, S3ErrorRetries, ref ParallelErrorNumber, ref ParallelErrorDescription, ref ParallelLogData, ref ParallelRequestHeaders, ref ParallelResponseStatusCode, ref ParallelResponseStatusDescription, ref ParallelResponseHeaders, ref ParallelResponseString);

                            //WriteToLogStream("Request=" + XMLItem, 2);
                            //WriteToLogStream("Response=" + ParallelResponseString, 2);

                            if (ParallelRetBool == true)
                            {
                                //go through the response and see is any errors

                                System.Xml.XmlDocument MyXmlDocument;
                                System.Xml.XmlNamespaceManager MyXmlNamespaceManager;
                                System.Xml.XmlNodeList MyXmlNodeList;
                                System.Xml.XmlNode KeyXmlNode;
                                System.Xml.XmlNode MessageXmlNode;

                                MyXmlDocument = new System.Xml.XmlDocument();
                                MyXmlDocument.LoadXml(ParallelResponseString);

                                MyXmlNamespaceManager = new System.Xml.XmlNamespaceManager(MyXmlDocument.NameTable);
                                MyXmlNamespaceManager.AddNamespace("amz", "http://s3.amazonaws.com/doc/2006-03-01/");

                                MyXmlNodeList = MyXmlDocument.SelectNodes("amz:DeleteResult/amz:Deleted", MyXmlNamespaceManager);

                                foreach (System.Xml.XmlNode ItemXmlNode in MyXmlNodeList)
                                {
                                    WriteToLogStream("   Delete successful.   KeyName=" + ItemXmlNode.SelectSingleNode("amz:Key", MyXmlNamespaceManager).InnerText, 2);
                                    DeleteCounter += 1;
                                }


                                MyXmlNodeList = MyXmlDocument.SelectNodes("amz:DeleteResult/amz:Error", MyXmlNamespaceManager);

                                foreach (System.Xml.XmlNode ItemXmlNode in MyXmlNodeList)
                                {
                                    WriteToLogStream("   Delete failed.   KeyName=" + ItemXmlNode.SelectSingleNode("amz:Key", MyXmlNamespaceManager).InnerText + "   Message=" + ItemXmlNode.SelectSingleNode("amz:Message", MyXmlNamespaceManager).InnerText);
                                    ExitCode = -1;
                                }

                            }
                            else
                            {
                                MyParallelLoopState.Stop();

                                WriteToLogStream(ParallelLogData, 2);
                                ParallelResponseErrorMessage = FormatErrorData(ParallelRequestURL, ParallelRequestMethod, ParallelRequestHeaders, ParallelResponseStatusCode, ParallelResponseStatusDescription, ParallelResponseHeaders, ParallelResponseString, ParallelErrorNumber, ParallelErrorDescription);
                                WriteToLogStream(ParallelResponseErrorMessage);

                                ExitCode = -1;
                            }
                        });

                        if (LogFilePath != "")
                        {
                            LogStreamWriter.Close();
                        }

                        WriteToLog("Number of items deleted: " + DeleteCounter);

                    }
                }
            }

            SendEmailLog();

            Environment.ExitCode = ExitCode;

            WriteToLog("");
            WriteToLog("Program complete.   " + DateTime.Now + "   ExitCode=" + ExitCode);
        }


        private static Boolean PopulateS3DeleteList(String AWSAccessKeyId, String AWSSecretAccessKey, Boolean UseSSL, String RequestEndpoint, String BucketName, String S3KeyName, int RetryTimes, ref System.Collections.Generic.List<String> S3DeleteList)
        {
            SprightlySoftAWS.REST MyREST = new SprightlySoftAWS.REST();
            String RequestURL;
            String RequestMethod;
            Boolean RetBool;
            int ErrorNumber = 0;
            String ErrorDescription = "";
            String LogData = "";
            Dictionary<String, String> RequestHeaders = new Dictionary<String, String>();
            int ResponseStatusCode = 0;
            String ResponseStatusDescription = "";
            Dictionary<String, String> ResponseHeaders = new Dictionary<String, String>();
            String ResponseString = "";
            String ResponseErrorMessage = "";

            String Marker = "";
            int MaxKeys = 1000;
            String QueryString;
            Boolean IsTruncated = true;
            String KeyName;
            DateTime LastModified;
            String LookupDelimiter = "";
            String DiffKeyName;

            System.Xml.XmlDocument MyXmlDocument;
            System.Xml.XmlNamespaceManager MyXmlNamespaceManager;
            System.Xml.XmlNode MyXmlNode;
            System.Xml.XmlNodeList MyXmlNodeList;

            //Add the root folder to the list;
            if (DeleteSpecifiedFolder == true)
            {
                if (S3KeyName != "")
                {
                    S3DeleteList.Add(S3KeyName);
                }
            }

            do
            {
                QueryString = "?delimiter=" + System.Uri.EscapeDataString(LookupDelimiter) + "&prefix=" + System.Uri.EscapeDataString(S3KeyName) + "&marker=" + System.Uri.EscapeDataString(Marker) + "&max-keys=" + MaxKeys;
                RequestURL = MyREST.BuildS3RequestURL(UseSSL, RequestEndpoint, BucketName, "", QueryString);
                RequestMethod = "GET";

                RetBool = MakeS3Request(AWSAccessKeyId, AWSSecretAccessKey, RequestURL, RequestMethod, null, "", RetryTimes, ref ErrorNumber, ref ErrorDescription, ref LogData, ref RequestHeaders, ref ResponseStatusCode, ref ResponseStatusDescription, ref ResponseHeaders, ref ResponseString);

                if (RetBool == true)
                {
                    MyXmlDocument = new System.Xml.XmlDocument();
                    MyXmlDocument.LoadXml(ResponseString);

                    MyXmlNamespaceManager = new System.Xml.XmlNamespaceManager(MyXmlDocument.NameTable);
                    MyXmlNamespaceManager.AddNamespace("amz", "http://s3.amazonaws.com/doc/2006-03-01/");

                    MyXmlNode = MyXmlDocument.SelectSingleNode("/amz:ListBucketResult/amz:NextMarker", MyXmlNamespaceManager);
                    if (MyXmlNode != null)
                    {
                        Marker = MyXmlNode.InnerText;
                    }

                    MyXmlNode = MyXmlDocument.SelectSingleNode("/amz:ListBucketResult/amz:IsTruncated", MyXmlNamespaceManager);
                    IsTruncated = Convert.ToBoolean(MyXmlNode.InnerText);

                    MyXmlNodeList = MyXmlDocument.SelectNodes("/amz:ListBucketResult/amz:Contents", MyXmlNamespaceManager);

                    foreach (System.Xml.XmlNode ItemXmlNode in MyXmlNodeList)
                    {
                        MyXmlNode = ItemXmlNode.SelectSingleNode("amz:Key", MyXmlNamespaceManager);
                        KeyName = MyXmlNode.InnerText;

                        MyXmlNode = ItemXmlNode.SelectSingleNode("amz:LastModified", MyXmlNamespaceManager);
                        LastModified = Convert.ToDateTime(MyXmlNode.InnerText);


                        if (KeyName != S3KeyName)
                        {
                            DiffKeyName = KeyName.Substring(S3KeyName.Length);

                            if (DeleteSubFolders == false && DiffKeyName.Contains(S3FolderDelimiter) == true)
                            {
                                WriteToLog("   S3 item not included because DeleteSubFolders=false.   KeyName=" + KeyName, 2);
                            }
                            else
                            {
                                if (KeyName.EndsWith(S3FolderDelimiter) == true)
                                {
                                    WriteToLog("   Folder added to S3 list.   KeyName=" + KeyName, 2);
                                    S3DeleteList.Add(KeyName);
                                }
                                else
                                {
                                    if (ExcludeS3FilesRegularExpression == "")
                                    {
                                        if (IncludeOnlyS3FilesRegularExpression == "")
                                        {
                                            if (DeleteFilesOlderThanDays > 0)
                                            {
                                                TimeSpan MyTimeSpan;
                                                MyTimeSpan = DateTime.Now - LastModified;

                                                if (MyTimeSpan.TotalDays > DeleteFilesOlderThanDays)
                                                {
                                                    WriteToLog("   File added to S3 list.   KeyName=" + KeyName, 2);
                                                    S3DeleteList.Add(KeyName);
                                                }
                                                else
                                                {
                                                    WriteToLog("   S3 file not included by DeleteFilesOlderThanDays.   KeyName=" + KeyName, 2);
                                                    RemoveParentFoldersFromList(KeyName, ref S3DeleteList);
                                                }
                                            }
                                            else
                                            {
                                                WriteToLog("   File added to S3 list.   KeyName=" + KeyName, 2);
                                                S3DeleteList.Add(KeyName);
                                            }
                                        }
                                        else
                                        {
                                            if (System.Text.RegularExpressions.Regex.IsMatch(KeyName, IncludeOnlyS3FilesRegularExpression, System.Text.RegularExpressions.RegexOptions.IgnoreCase) == true)
                                            {
                                                if (DeleteFilesOlderThanDays > 0)
                                                {
                                                    TimeSpan MyTimeSpan;
                                                    MyTimeSpan = DateTime.Now - LastModified;

                                                    if (MyTimeSpan.TotalDays > DeleteFilesOlderThanDays)
                                                    {
                                                        WriteToLog("   File added to S3 list.   KeyName=" + KeyName, 2);
                                                        S3DeleteList.Add(KeyName);
                                                    }
                                                    else
                                                    {
                                                        WriteToLog("   S3 file not included by DeleteFilesOlderThanDays.   KeyName=" + KeyName, 2);
                                                        RemoveParentFoldersFromList(KeyName, ref S3DeleteList);
                                                    }
                                                }
                                                else
                                                {
                                                    WriteToLog("   File added to S3 list.   KeyName=" + KeyName, 2);
                                                    S3DeleteList.Add(KeyName);
                                                }
                                            }
                                            else
                                            {
                                                WriteToLog("   S3 file not included by IncludeOnlyS3FilesRegularExpression.   KeyName=" + KeyName, 2);
                                                RemoveParentFoldersFromList(KeyName, ref S3DeleteList);
                                            }
                                        }
                                    }
                                    else
                                    {
                                        if (System.Text.RegularExpressions.Regex.IsMatch(KeyName, ExcludeS3FilesRegularExpression, System.Text.RegularExpressions.RegexOptions.IgnoreCase) == true)
                                        {
                                            WriteToLog("   S3 file excluded by ExcludeS3FilesRegularExpression.   KeyName=" + KeyName, 2);
                                            RemoveParentFoldersFromList(KeyName, ref S3DeleteList);
                                        }
                                        else
                                        {
                                            if (IncludeOnlyS3FilesRegularExpression == "")
                                            {
                                                if (DeleteFilesOlderThanDays > 0)
                                                {
                                                    TimeSpan MyTimeSpan;
                                                    MyTimeSpan = DateTime.Now - LastModified;

                                                    if (MyTimeSpan.TotalDays > DeleteFilesOlderThanDays)
                                                    {
                                                        WriteToLog("   File added to S3 list.   KeyName=" + KeyName, 2);
                                                        S3DeleteList.Add(KeyName);
                                                    }
                                                    else
                                                    {
                                                        WriteToLog("   S3 file not included by DeleteFilesOlderThanDays.   KeyName=" + KeyName, 2);
                                                        RemoveParentFoldersFromList(KeyName, ref S3DeleteList);
                                                    }
                                                }
                                                else
                                                {
                                                    WriteToLog("   File added to S3 list.   KeyName=" + KeyName, 2);
                                                    S3DeleteList.Add(KeyName);
                                                }
                                            }
                                            else
                                            {
                                                if (System.Text.RegularExpressions.Regex.IsMatch(KeyName, IncludeOnlyS3FilesRegularExpression, System.Text.RegularExpressions.RegexOptions.IgnoreCase) == true)
                                                {
                                                    if (DeleteFilesOlderThanDays > 0)
                                                    {
                                                        TimeSpan MyTimeSpan;
                                                        MyTimeSpan = DateTime.Now - LastModified;

                                                        if (MyTimeSpan.TotalDays > DeleteFilesOlderThanDays)
                                                        {
                                                            WriteToLog("   File added to S3 list.   KeyName=" + KeyName, 2);
                                                            S3DeleteList.Add(KeyName);
                                                        }
                                                        else
                                                        {
                                                            WriteToLog("   S3 file not included by DeleteFilesOlderThanDays.   KeyName=" + KeyName, 2);
                                                            RemoveParentFoldersFromList(KeyName, ref S3DeleteList);
                                                        }
                                                    }
                                                    else
                                                    {
                                                        WriteToLog("   File added to S3 list.   KeyName=" + KeyName, 2);
                                                        S3DeleteList.Add(KeyName);
                                                    }
                                                }
                                                else
                                                {
                                                    WriteToLog("   S3 file not included by IncludeOnlyS3FilesRegularExpression.   KeyName=" + KeyName, 2);
                                                    RemoveParentFoldersFromList(KeyName, ref S3DeleteList);
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }

                        if (LookupDelimiter == "")
                        {
                            Marker = KeyName;
                        }
                    }


                    MyXmlNodeList = MyXmlDocument.SelectNodes("/amz:ListBucketResult/amz:CommonPrefixes", MyXmlNamespaceManager);

                    //No prefixes will exist since we are using a empty delimiter.
                    foreach (System.Xml.XmlNode ItemXmlNode in MyXmlNodeList)
                    {
                        MyXmlNode = ItemXmlNode.SelectSingleNode("amz:Prefix", MyXmlNamespaceManager);
                        KeyName = MyXmlNode.InnerText;

                        WriteToLog("   Folder added to S3 list.   KeyName=" + KeyName, 2);
                        S3DeleteList.Add(KeyName);
                    }
                }
                else
                {
                    WriteToLog(LogData, 2);
                    ResponseErrorMessage = FormatErrorData(RequestURL, RequestMethod, RequestHeaders, ResponseStatusCode, ResponseStatusDescription, ResponseHeaders, ResponseString, ErrorNumber, ErrorDescription);
                    WriteToLog("   Error listing files on S3.   " + ResponseErrorMessage);
                    ExitCode = -1;
                    break;
                }

            } while (IsTruncated == true);

            return RetBool;
        }


        private static void RemoveParentFoldersFromList(String KeyName, ref System.Collections.Generic.List<String> S3DeleteList)
        {
            String ParentFolder = KeyName;

            while (ParentFolder.Contains(S3FolderDelimiter))
            {
                ParentFolder = ParentFolder.Substring(0, ParentFolder.LastIndexOf(S3FolderDelimiter) + 1);
                S3DeleteList.Remove(ParentFolder);
                ParentFolder = ParentFolder.Substring(0, ParentFolder.LastIndexOf(S3FolderDelimiter));
            }
        }


        private static Boolean MakeS3Request(String AWSAccessKeyId, String AWSSecretAccessKey, String RequestURL, String RequestMethod, System.Collections.Generic.Dictionary<String, String> UserInputRequestHeaders, String SendData, int RetryTimes, ref int ErrorNumber, ref String ErrorDescription, ref String LogData, ref System.Collections.Generic.Dictionary<String, String> RequestHeaders, ref int ResponseStatusCode, ref String ResponseStatusDescription, ref System.Collections.Generic.Dictionary<String, String> ResponseHeaders, ref String ResponseStringFormatted)
        {
            SprightlySoftAWS.REST MyREST = new SprightlySoftAWS.REST();
            Boolean RetBool = true;
            LogData = "";

            Dictionary<String, String> FullInputRequestHeaders;
            String AuthorizationValue;

            for (int i = 0; i <= RetryTimes; i++)
            {
                FullInputRequestHeaders = new Dictionary<String, String>();

                if (UserInputRequestHeaders != null)
                {
                    foreach (KeyValuePair<String, String> MyKeyValuePair in UserInputRequestHeaders)
                    {
                        FullInputRequestHeaders.Add(MyKeyValuePair.Key, MyKeyValuePair.Value);
                    }
                }

                //Set the date for each request.
                FullInputRequestHeaders.Add("x-amz-date", DateTime.UtcNow.ToString("r"));

                AuthorizationValue = MyREST.GetS3AuthorizationValue(RequestURL, RequestMethod, FullInputRequestHeaders, AWSAccessKeyId, AWSSecretAccessKey);
                FullInputRequestHeaders.Add("Authorization", AuthorizationValue);

                RetBool = MyREST.MakeRequest(RequestURL, RequestMethod, FullInputRequestHeaders, SendData);

                //Set the return values.
                ErrorNumber = MyREST.ErrorNumber;
                ErrorDescription = MyREST.ErrorDescription;
                LogData += MyREST.LogData + Environment.NewLine;
                ResponseStatusCode = MyREST.ResponseStatusCode;
                ResponseStatusDescription = MyREST.ResponseStatusDescription;
                ResponseHeaders = MyREST.ResponseHeaders;
                ResponseStringFormatted = MyREST.ResponseStringFormatted;

                if (RetBool == true)
                {
                    break;
                }
                else
                {
                    if (MyREST.ResponseStatusCode == 500 || MyREST.ResponseStatusCode == 503)
                    {
                        //A Service Unavailable response was returned.  Wait and retry.
                        System.Threading.Thread.Sleep(1000 * i * i);
                    }
                    else if (MyREST.ResponseStatusCode == 0 && MyREST.ErrorNumber == 1003)
                    {
                        //Getting the response failed.  This may be a network disconnection.  Wait and retry.
                        System.Threading.Thread.Sleep(1000 * i * i);
                    }
                    else
                    {
                        //An error occured but retrying would not solve the problem.
                        break;
                    }
                }
            }

            return RetBool;
        }


        private static String RunCommand(String CommandFile, String CommandArguments)
        {
            System.Diagnostics.Process MyProcess;
            System.Diagnostics.ProcessStartInfo MyProcessStartInfo;
            System.IO.StreamReader MyStreamReader;
            System.IO.StreamReader MyStreamReaderError;
            String OutputLine = "";

            MyProcessStartInfo = new System.Diagnostics.ProcessStartInfo();
            MyProcessStartInfo.UseShellExecute = false;
            MyProcessStartInfo.RedirectStandardOutput = true;
            MyProcessStartInfo.RedirectStandardError = true;

            MyProcessStartInfo.FileName = CommandFile;
            MyProcessStartInfo.Arguments = CommandArguments;

            MyProcess = new System.Diagnostics.Process();
            MyProcess.StartInfo = MyProcessStartInfo;
            MyProcess.Start();

            MyStreamReader = MyProcess.StandardOutput;
            MyStreamReaderError = MyProcess.StandardError;

            if (MyStreamReader.EndOfStream == false)
            {
                while (MyStreamReader.EndOfStream == false)
                {
                    OutputLine += MyStreamReader.ReadLine() + Environment.NewLine;
                }
            }
            else
            {
                while (MyStreamReaderError.EndOfStream == false)
                {
                    OutputLine += MyStreamReaderError.ReadLine() + Environment.NewLine;
                }
            }

            MyProcess.WaitForExit();
            MyProcess.Close();

            return OutputLine;
        }


        private static String GetValueFromConfigOrArgument(string[] args, String ArgumentName)
        {
            String CurrentValue;
            CurrentValue = System.Configuration.ConfigurationManager.AppSettings[ArgumentName];

            ArgumentName = ArgumentName.ToLower();
            String CurrentArg;

            for (int i = 0; i < args.Length; i++)
            {
                CurrentArg = args[i].ToLower();

                if (CurrentArg.StartsWith("-" + ArgumentName) | CurrentArg.StartsWith("/" + ArgumentName) | CurrentArg.StartsWith("–" + ArgumentName))
                {
                    try
                    {
                        CurrentValue = args[i + 1];
                    }
                    catch (Exception ex)
                    {
                        WriteToLog("Error reading command line value.  ArgumentName=" + ArgumentName + "   ex.Message=" + ex.Message);
                    }

                    break;
                }
            }

            if (CurrentValue == null)
            {
                CurrentValue = "";
            }

            return CurrentValue;
        }


        private static void SendEmailLog()
        {
            //Send the log information in an email.
            if ((EmailLogCondition == 1 & ExitCode != 0) | EmailLogCondition == 2)
            {
                String EmailSubject;

                if (ExitCode == 0)
                {
                    EmailSubject = "S3Delete Success";
                }
                else
                {
                    EmailSubject = "S3Delete Failed";
                }

                WriteToLog("");
                WriteToLog("Sending the log email.");
                SendEmail(EmailSmtpHost, EmailSmtpPort, EmailSmtpEnableSsl, EmailSmtpUserName, EmailSmtpPassword, EmailFromAddress, EmailRecipients, EmailSubject, EmailLogData);
            }
        }


        private static void SendEmail(String EmailSmtpHost, int EmailSmtpPort, Boolean EmailSmtpEnableSsl, String EmailSmtpUserName, String EmailSmtpPassword, String EmailFromAddress, String EmailRecipients, String EmailSubject, String EmailBody)
        {
            try
            {
                System.Net.Mail.SmtpClient MySmtpClient = new System.Net.Mail.SmtpClient();
                MySmtpClient.Host = EmailSmtpHost;
                MySmtpClient.Port = EmailSmtpPort;
                MySmtpClient.EnableSsl = EmailSmtpEnableSsl;

                if (EmailSmtpUserName != "")
                {
                    System.Net.NetworkCredential MyNetworkCredential = new System.Net.NetworkCredential();
                    MyNetworkCredential.UserName = EmailSmtpUserName;
                    MyNetworkCredential.Password = EmailSmtpPassword;

                    MySmtpClient.Credentials = MyNetworkCredential;
                }

                System.Net.Mail.MailAddress FromMailAddress = new System.Net.Mail.MailAddress(EmailFromAddress);
                System.Net.Mail.MailAddress ToMailAddress = new System.Net.Mail.MailAddress(EmailRecipients);

                System.Net.Mail.MailMessage MyMailMessage = new System.Net.Mail.MailMessage(FromMailAddress, ToMailAddress);

                MyMailMessage.Subject = EmailSubject;
                MyMailMessage.Body = EmailBody;

                MySmtpClient.Send(MyMailMessage);
            }
            catch (Exception ex)
            {
                WriteToLog("Sending the log email failed.   ex.Message=" + ex.Message);
                ExitCode = -1;
            }
        }


        private static void InitializeLogFile(Boolean OverwriteExistingLogFile)
        {
            if (LogFilePath != "")
            {
                String LogLine = "";

                if (OverwriteExistingLogFile == true)
                {
                    try
                    {
                        if (System.IO.File.Exists(LogFilePath) == true)
                        {
                            LogLine = "Deleting the existing log file.   LogFilePath=" + LogFilePath;
                            System.IO.File.Delete(LogFilePath);
                        }
                    }
                    catch (Exception ex)
                    {
                    }
                }

                try
                {
                    System.IO.StreamWriter MyStreamWriter;
                    MyStreamWriter = new System.IO.StreamWriter(LogFilePath, true, System.Text.Encoding.Unicode);
                    MyStreamWriter.Close();

                    if (LogLine != "")
                    {
                        WriteToLog(LogLine);
                    }
                }
                catch (Exception ex)
                {
                    LogLine = "Error opening log file.  No data will be written to the log file.   LogFilePath=" + LogFilePath + "   ex.Message=" + ex.Message;
                    Console.WriteLine(LogLine);
                    EmailLogData += LogLine + Environment.NewLine;
                    LogFilePath = "";
                }
            }
        }


        private static String FormatErrorData(String RequestURL, String RequestMethod, Dictionary<String, String> RequestHeaders, int ResponseStatusCode, String ResponseStatusDescription, Dictionary<String, String> ResponseHeaders, String ResponseString, int ErrorNumber, String ErrorDescription)
        {
            String ReturnString = "";
            ReturnString += "An error occurred." + Environment.NewLine;

            if (OutputLevel == 1)
            {
                if (ResponseString == "")
                {
                    ReturnString += "Component ErrorDescription: " + ErrorDescription + Environment.NewLine;
                    ReturnString += "Component ErrorNumber: " + ErrorNumber + Environment.NewLine;
                    ReturnString += "Response Status Code: " + ResponseStatusCode + Environment.NewLine;
                    ReturnString += "Response Status Description: " + ResponseStatusDescription;
                }
                else
                {
                    try
                    {
                        System.Xml.XmlDocument XmlDoc = new System.Xml.XmlDocument();
                        XmlDoc.LoadXml(ResponseString);

                        System.Xml.XmlNode XmlNode;
                        XmlNode = XmlDoc.SelectSingleNode("/Error/Message");

                        if (XmlNode == null)
                        {
                            ReturnString += "Response XML: " + Environment.NewLine + ResponseString;
                        }
                        else
                        {
                            ReturnString += "   Amazon Error Message: " + XmlNode.InnerText;
                        }
                    }
                    catch (Exception e)
                    {
                        ReturnString += "Response String: " + Environment.NewLine + ResponseString;
                    }
                }
            }
            else
            {
                if (ErrorNumber != 0)
                {
                    ReturnString += "Component ErrorDescription: " + ErrorDescription + Environment.NewLine;
                    ReturnString += "Component ErrorNumber: " + ErrorNumber + Environment.NewLine;
                    ReturnString += Environment.NewLine;
                }

                if (ResponseStatusCode != 0)
                {
                    ReturnString += "Request URL: " + RequestURL + Environment.NewLine;
                    ReturnString += "Request Method: " + RequestMethod + Environment.NewLine;

                    foreach (KeyValuePair<String, String> MyHeader in RequestHeaders)
                    {
                        ReturnString += "Request Header: " + MyHeader.Key + ":" + MyHeader.Value + Environment.NewLine;
                    }

                    ReturnString += Environment.NewLine;
                    ReturnString += "Response Status Code: " + ResponseStatusCode + Environment.NewLine;
                    ReturnString += "Response Status Description: " + ResponseStatusDescription + Environment.NewLine;

                    foreach (KeyValuePair<String, String> MyHeader in ResponseHeaders)
                    {
                        ReturnString += "Response Header: " + MyHeader.Key + ":" + MyHeader.Value + Environment.NewLine;
                    }

                    if (ResponseString != "")
                    {
                        ReturnString += Environment.NewLine;

                        try
                        {
                            System.Xml.XmlDocument XmlDoc = new System.Xml.XmlDocument();
                            XmlDoc.LoadXml(ResponseString);

                            ReturnString += "Response XML: " + Environment.NewLine + ResponseString + Environment.NewLine;
                        }
                        catch (Exception e)
                        {
                            ReturnString += "Response String: " + Environment.NewLine + ResponseString + Environment.NewLine;
                        }
                    }
                }
            }

            return ReturnString;
        }


        private static void WriteToLog(String LogLine, int LogLineLevel = 1)
        {
            if (LogLineLevel <= OutputLevel)
            {
                Console.WriteLine(LogLine);
                System.Diagnostics.Debug.Print(LogLine);

                if (OutputLevel > 1)
                {
                    LogLine = DateTime.Now + "   " + LogLine;
                }

                if (LogFilePath != "")
                {
                    System.IO.StreamWriter MyStreamWriter;
                    MyStreamWriter = new System.IO.StreamWriter(LogFilePath, true, System.Text.Encoding.Unicode);
                    MyStreamWriter.WriteLine(LogLine);
                    MyStreamWriter.Close();
                }

                EmailLogData += LogLine + Environment.NewLine;
            }
        }


        private static void WriteToLogStream(String LogLine, int LogLineLevel = 1)
        {
            if (LogLineLevel <= OutputLevel)
            {
                Console.WriteLine(LogLine);
                System.Diagnostics.Debug.Print(LogLine);

                if (OutputLevel > 1)
                {
                    LogLine = DateTime.Now + "   " + LogLine;
                }

                if (LogFilePath != "")
                {
                    LogStreamWriter.WriteLine(LogLine);
                    LogStreamWriter.Flush();
                }

                EmailLogData += LogLine + Environment.NewLine;
            }
        }


        static void MyUnhandledException(object sender, UnhandledExceptionEventArgs args)
        {
            Exception e = (Exception)args.ExceptionObject;

            WriteToLog("");
            WriteToLog("Error in application.");
            WriteToLog("Message=" + e.Message);
            WriteToLog("StackTrace=" + e.StackTrace);

            ExitCode = -1;
            Environment.ExitCode = ExitCode;

            //Send the log email.
            SendEmailLog();
        }

    }

    class BucketItemObject
    {
        public long Size;
        public String ETag;
    }

}
