﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;

namespace S3UploadDownload
{
    public partial class FormDownloadEstimate : Form
    {
        SprightlySoftAWS.S3.CalculateHash MyCalculateHash;
        SprightlySoftAWS.S3.DownloadEstimate MyDownload;
        System.ComponentModel.BackgroundWorker CalculateHashBackgroundWorker;
        System.ComponentModel.BackgroundWorker DownloadBackgroundWorker;
        System.ComponentModel.BackgroundWorker ETagBackgroundWorker;
        System.Timers.Timer TimerUpdateStatus;
        FormConflictDownload MyFormConflictDownload;
        Double ExistingFileSize;

        public FormDownloadEstimate()
        {
            InitializeComponent();
        }

        private void FormDownloadEstimate_Load(object sender, EventArgs e)
        {
            Properties.Settings MySettings = new Properties.Settings();

            TextBoxDownloadBucketName.Text = MySettings.DownloadEstimateBucketName;
            TextBoxDownloadKeyName.Text = MySettings.DownloadEstimateKeyName;
            TextBoxDownloadFileName.Text = MySettings.DownloadEstimateFileName;

            MyCalculateHash = new SprightlySoftAWS.S3.CalculateHash();

            MyDownload = new SprightlySoftAWS.S3.DownloadEstimate();
            MyDownload.ProgressChangedEvent += MyDownload_ProgressChangedEvent;

            DownloadBackgroundWorker = new System.ComponentModel.BackgroundWorker();
            DownloadBackgroundWorker.DoWork += DownloadBackgroundWorker_DoWork;
            DownloadBackgroundWorker.RunWorkerCompleted += DownloadBackgroundWorker_RunWorkerCompleted;

            ETagBackgroundWorker = new System.ComponentModel.BackgroundWorker();
            ETagBackgroundWorker.DoWork += ETagBackgroundWorker_DoWork;
            ETagBackgroundWorker.RunWorkerCompleted += ETagBackgroundWorker_RunWorkerCompleted;

            //use a timer to show the current speed and time remaining
            TimerUpdateStatus = new System.Timers.Timer();
            TimerUpdateStatus.Elapsed += new System.Timers.ElapsedEventHandler(TimerUpdateStatus_Tick);
            TimerUpdateStatus.Interval = 500;

        }

        protected override void OnFormClosing(FormClosingEventArgs e)
        {
            Properties.Settings MySettings = new Properties.Settings();

            MySettings.DownloadEstimateBucketName = TextBoxDownloadBucketName.Text;
            MySettings.DownloadEstimateKeyName = TextBoxDownloadKeyName.Text;
            MySettings.DownloadEstimateFileName = TextBoxDownloadFileName.Text;

            MySettings.Save();
        }

        private void EnableDisableStart()
        {
            TextBoxDownloadBucketName.Enabled = false;
            TextBoxDownloadKeyName.Enabled = false;
            TextBoxDownloadFileName.Enabled = false;
            ButtonDownloadFile.Enabled = false;
            ButtonAbort.Enabled = true;

            ToolStripProgressBar.Value = 0;
            ToolStripStatusLabelBytes.Text = "";
            ToolStripStatusLabelSpeed.Text = "";
            ToolStripStatusLabelElapsed.Text = "";
            ToolStripStatusLabelRemaining.Text = "";

            ToolStripProgressBar.Visible = true;
            ToolStripStatusLabelBytes.Visible = true;
            ToolStripStatusLabelSpeed.Visible = true;
            ToolStripStatusLabelElapsed.Visible = true;
            ToolStripStatusLabelRemaining.Visible = true;

            TimerUpdateStatus.Enabled = true;
        }

        private void EnableDisableEnd()
        {
            TextBoxDownloadBucketName.Enabled = true;
            TextBoxDownloadKeyName.Enabled = true;
            TextBoxDownloadFileName.Enabled = true;
            ButtonDownloadFile.Enabled = true;

            ButtonAbort.Enabled = false;

            ToolStripProgressBar.Visible = false;
            ToolStripStatusLabelBytes.Visible = false;
            ToolStripStatusLabelSpeed.Visible = false;
            ToolStripStatusLabelElapsed.Visible = false;
            ToolStripStatusLabelRemaining.Visible = false;

            TimerUpdateStatus.Enabled = false;
        }

        private void ButtonDownloadFile_Click(object sender, EventArgs e)
        {
            EnableDisableStart();

            if (TextBoxDownloadFileName.Text.Contains("\\") == true && System.IO.Directory.Exists(System.IO.Path.GetDirectoryName(TextBoxDownloadFileName.Text)) == true)
            {
                ExistingFileSize = 0;
                Dictionary<String, String> ExtraRequestHeaders = new Dictionary<String, String>();
                Boolean AppendToFile = false;
                Boolean SkipDownload = false;
                String RequestURL;
                String RequestMethod;
                String AuthorizationValue;

                if (System.IO.File.Exists(TextBoxDownloadFileName.Text) == true)
                {
                    //The file to download already exists.  Populate the conflict form and ask
                    //the user what to do.

                    System.IO.FileInfo LocalFileInfo= new System.IO.FileInfo(TextBoxDownloadFileName.Text);
                    ExistingFileSize = LocalFileInfo.Length;
                    MyFormConflictDownload = new FormConflictDownload();
                    MyFormConflictDownload.TextBoxLocalFileName.Text = TextBoxDownloadFileName.Text;
                    MyFormConflictDownload.TextBoxLocalSize.Text = LocalFileInfo.Length.ToString("N0") + " bytes";
                    MyFormConflictDownload.TextBoxLocalDate.Text = Convert.ToString(LocalFileInfo.LastWriteTime);

                    //kick off a process to calculate the etag.  A large file may take a long time to calculate.
                    MyFormConflictDownload.TextBoxLocalETag.Text = "calculating....";
                    System.Collections.Hashtable ETagHashTable = new System.Collections.Hashtable();
                    ETagHashTable.Add("LocalFileName", TextBoxDownloadFileName.Text);
                    ETagBackgroundWorker.RunWorkerAsync(ETagHashTable);

                    Properties.Settings MySettings = new Properties.Settings();

                    //get the properties of the Amazon file
                    SprightlySoftAWS.REST MyREST = new SprightlySoftAWS.REST();

                    RequestURL = MyDownload.BuildS3RequestURL(true, "s3.amazonaws.com", TextBoxDownloadBucketName.Text, TextBoxDownloadKeyName.Text, "");

                    RequestMethod = "HEAD";

                    ExtraRequestHeaders = new Dictionary<String, String>();
                    ExtraRequestHeaders.Add("x-amz-date", DateTime.UtcNow.ToString("r"));

                    AuthorizationValue = MyDownload.GetS3AuthorizationValue(RequestURL, RequestMethod, ExtraRequestHeaders, MySettings.AWSAccessKeyId, MySettings.AWSSecretAccessKey);
                    ExtraRequestHeaders.Add("Authorization", AuthorizationValue);

                    Boolean RetBool;
                    //RetBool = MyOperation.GetOperation(MySettings.AWSAccessKeyId, MySettings.AWSSecretAccessKey, TextBoxDownloadBucketName.Text, TextBoxDownloadKeyName.Text, "", "HEAD", null);
                    RetBool = MyREST.MakeRequest(RequestURL, RequestMethod, ExtraRequestHeaders, "");

                    if (RetBool == true)
                    {
                        MyFormConflictDownload.TextBoxRemoteBucketName.Text = TextBoxDownloadBucketName.Text;
                        MyFormConflictDownload.TextBoxRemoteKeyName.Text = TextBoxDownloadKeyName.Text;
                        MyFormConflictDownload.TextBoxRemoteSize.Text = Convert.ToDouble(MyREST.ResponseHeaders["Content-Length"]).ToString("N0") + " bytes";
                        MyFormConflictDownload.TextBoxRemoteAmazonDate.Text = Convert.ToString(Convert.ToDateTime(MyREST.ResponseHeaders["Last-Modified"]));
                        MyFormConflictDownload.TextBoxRemoteETag.Text = MyREST.ResponseHeaders["ETag"].Replace("\"", "");
                    }
                    else
                    {
                        EnableDisableEnd();
                        MessageBox.Show(MyREST.ErrorDescription, "Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
                        return;
                    }

                    //only allow resume if the Amazon file is larger
                    if (Convert.ToDouble(MyREST.ResponseHeaders["Content-Length"]) > LocalFileInfo.Length)
                    {
                        MyFormConflictDownload.ButtonResume.Enabled = true;
                    }
                    else
                    {
                        MyFormConflictDownload.ButtonResume.Enabled = false;
                    }

                    //show the conflict form
                    System.Windows.Forms.DialogResult DialogResult;
                    DialogResult = MyFormConflictDownload.ShowDialog(this);

                    if (DialogResult == System.Windows.Forms.DialogResult.Retry)
                    {
                        //only download the remainder of the file.   Add a range header to the request.
                        ExtraRequestHeaders.Add("Range", "bytes=" + LocalFileInfo.Length + "-" + (Convert.ToDouble(MyREST.ResponseHeaders["Content-Length"]) - 1));
                        AppendToFile = true;
                    }
                    else if (DialogResult == System.Windows.Forms.DialogResult.Yes)
                    {
                        //do nothing
                    }
                    else
                    {
                        SkipDownload = true;
                    }
                }

                if (SkipDownload == true)
                {
                    EnableDisableEnd();
                    MessageBox.Show("Download skipped.", "Skipped", MessageBoxButtons.OK, MessageBoxIcon.Information);
                }
                else
                {
                    Properties.Settings MySettings = new Properties.Settings();

                    RequestURL = MyDownload.BuildS3RequestURL(true, "s3.amazonaws.com", TextBoxDownloadBucketName.Text, TextBoxDownloadKeyName.Text, "");

                    RequestMethod = "GET";

                    ExtraRequestHeaders = new Dictionary<String, String>();
                    ExtraRequestHeaders.Add("x-amz-date", DateTime.UtcNow.ToString("r"));

                    AuthorizationValue = MyDownload.GetS3AuthorizationValue(RequestURL, RequestMethod, ExtraRequestHeaders, MySettings.AWSAccessKeyId, MySettings.AWSSecretAccessKey);
                    ExtraRequestHeaders.Add("Authorization", AuthorizationValue);

                    //Use a hash table to pass parameters to the function in the BackgroundWorker.
                    System.Collections.Hashtable DownloadHashTable = new System.Collections.Hashtable();
                    DownloadHashTable.Add("RequestURL", RequestURL);
                    DownloadHashTable.Add("RequestMethod", RequestMethod);
                    DownloadHashTable.Add("ExtraRequestHeaders", ExtraRequestHeaders);
                    DownloadHashTable.Add("LocalFileName", TextBoxDownloadFileName.Text);
                    DownloadHashTable.Add("AppendToFile", AppendToFile);

                    //Run the DownloadFile function in a BackgroundWorker process.  Downloading a large
                    //file may take a long time.  Running the process in a BackgroundWorker will prevent
                    //the Window from locking up.
                    DownloadBackgroundWorker.RunWorkerAsync(DownloadHashTable);
                }

            }
            else
            {
                EnableDisableEnd();
                MessageBox.Show("The local file path does not exist.", "Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
        }

        private void DownloadBackgroundWorker_DoWork(object sender, DoWorkEventArgs e)
        {
            //Call the DownloadFile function and set the result.  When the function is complete
            //the RunWorkerCompleted event will fire.  Use the parameters that were passed in the 
            //hash table.
            System.Collections.Hashtable DownloadHashTable = e.Argument as System.Collections.Hashtable;

            e.Result = MyDownload.DownloadFile(DownloadHashTable["RequestURL"].ToString(), DownloadHashTable["RequestMethod"].ToString(), DownloadHashTable["ExtraRequestHeaders"] as Dictionary<String, String>, DownloadHashTable["LocalFileName"].ToString(), Convert.ToBoolean(DownloadHashTable["AppendToFile"]));
        }

        private void DownloadBackgroundWorker_RunWorkerCompleted(object sender, System.ComponentModel.RunWorkerCompletedEventArgs e)
        {
            System.Diagnostics.Debug.Print("");
            System.Diagnostics.Debug.Print(MyDownload.LogData);
            System.Diagnostics.Debug.Print("");

            EnableDisableEnd();

            if (Convert.ToBoolean(e.Result) == true)
            {
                MessageBox.Show("Download complete.", "Success", MessageBoxButtons.OK, MessageBoxIcon.Information);
            }
            else
            {
                //Show the error message.
                String ResponseMessage;

                if (MyDownload.ResponseString == "")
                {
                    ResponseMessage = MyDownload.ErrorDescription;
                }
                else
                {
                    System.Xml.XmlDocument XmlDoc = new System.Xml.XmlDocument();
                    XmlDoc.LoadXml(MyDownload.ResponseString);

                    System.Xml.XmlNode XmlNode;
                    XmlNode = XmlDoc.SelectSingleNode("/Error/Message");

                    ResponseMessage = XmlNode.InnerText;
                }

                MessageBox.Show(ResponseMessage, "Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
        }

        private void ButtonAbort_Click(object sender, EventArgs e)
        {
            MyDownload.Abort();
        }

        private void MyDownload_ProgressChangedEvent()
        {
            if (this.InvokeRequired == true)
            {
                this.BeginInvoke(new MethodInvoker(delegate() { MyDownload_ProgressChangedEvent(); }));
            }
            else
            {
                //Update the progress bar when the ProgressChangedEvent is fired.
                if (MyDownload.BytesTotal > 0)
                {
                    decimal MyDecimal = (Convert.ToDecimal(MyDownload.BytesTransfered) / Convert.ToDecimal(MyDownload.BytesTotal)) * 100;
                    ToolStripProgressBar.Value = Convert.ToInt32(MyDecimal);

                    SprightlySoftAWS.S3.Helper MyHelper = new SprightlySoftAWS.S3.Helper();
                    ToolStripStatusLabelBytes.Text = MyHelper.FormatByteSize(MyDownload.BytesTransfered) + " / " + MyHelper.FormatByteSize(MyDownload.BytesTotal);
                }
            }
        }

        private void TimerUpdateStatus_Tick(object source, System.Timers.ElapsedEventArgs e)
        {
            if (this.InvokeRequired == true)
            {
                this.BeginInvoke(new MethodInvoker(delegate() { TimerUpdateStatus_Tick(source, e); }));
            }
            else
            {
                //get the current speed, time elapsed and time remaining from the DownloadEstimate class
                ToolStripStatusLabelSpeed.Text = MyDownload.CurrentKBpsSpeed.ToString("N0") + " KB/s";

                System.TimeSpan ElapsedTimeSpan;
                ElapsedTimeSpan = new System.TimeSpan(0, 0, Convert.ToInt32(MyDownload.SecondsElapsed));
                ToolStripStatusLabelElapsed.Text = ElapsedTimeSpan.ToString() + " elapsed";

                System.TimeSpan RemainingTimeSpan;
                RemainingTimeSpan = new System.TimeSpan(0, 0, Convert.ToInt32(MyDownload.SecondsRemaining));
                ToolStripStatusLabelRemaining.Text = RemainingTimeSpan.ToString() + " remaining";
            }
        }

        private void ETagBackgroundWorker_DoWork(object sender, DoWorkEventArgs e)
        {
            System.Collections.Hashtable CalculateHashHashTable = e.Argument as System.Collections.Hashtable;
            e.Result = MyCalculateHash.CalculateETagFromFile(CalculateHashHashTable["LocalFileName"].ToString());
        }

        private void ETagBackgroundWorker_RunWorkerCompleted(object sender, System.ComponentModel.RunWorkerCompletedEventArgs e)
        {
            //show the etag in the conflict form
            MyFormConflictDownload.TextBoxLocalETag.Text = e.Result.ToString();
        }

    }
}
